import ScoreImportFatalError from "lib/score-import/framework/score-importing/score-import-error";
import { ServerConfig } from "lib/setup/config";
import p from "prudence";
import nodeFetch from "utils/fetch";
import { URLSearchParams } from "url";
import type { CGErrorResponse, CGScoresResponse, CGServices, CGSupportedGames } from "./types";
import type { KtLogger } from "lib/logger/logger";
import type { PrudenceSchema } from "prudence";
import type { CGCardInfo } from "tachi-common";
import type { NodeFetch } from "utils/fetch";

const PR_CG_RESPONSE: PrudenceSchema = {
	success: "boolean",
	data: {
		profile: {
			id: "string",
			name: "string",
		},

		// validated elsewhere
		scores: [p.any],
	},
};

const PR_CG_ERR_RESPONSE: PrudenceSchema = {
	success: "boolean",
	message: "string",
};

/**
 * Fetch this info from CG. **This function does not validate the content of scores**,
 * instead, that should be handled by the parser.
 */
export async function FetchCGScores(
	service: CGServices,
	cardInfo: CGCardInfo,
	game: CGSupportedGames,
	logger: KtLogger,
	fetch: NodeFetch = nodeFetch
): Promise<Array<unknown>> {
	const url = GetCGUrl(service, cardInfo, game);

	let validatedRes: CGErrorResponse | CGScoresResponse<unknown>;

	try {
		const res: unknown = await fetch(url).then((r) => r.json());

		const prErr = p({ res }, { res: p.or(PR_CG_RESPONSE, PR_CG_ERR_RESPONSE) });

		if (prErr) {
			logger.error(`Got unexpected data from CG.`, { res });
			throw new Error(`Got unexpected data from CG.`);
		}

		validatedRes = res as CGErrorResponse | CGScoresResponse<unknown>;
	} catch (err) {
		logger.error(`Received invalid response from ${url}.`, { err });

		throw new ScoreImportFatalError(
			500,
			`Received invalid response from their API. Are they down?`
		);
	}

	if (validatedRes.success) {
		return validatedRes.data.scores;
	}

	throw new ScoreImportFatalError(
		400,
		`Failed to fetch data from CG. Error Message: ${validatedRes.message}.`
	);
}

function GetCGConf(service: CGServices) {
	switch (service) {
		case "dev": {
			if (!ServerConfig.CG_DEV_CONFIG) {
				throw new Error(
					`No CG_DEV_CONFIG is available, yet a request to cg-dev was attempted?`
				);
			}

			return ServerConfig.CG_DEV_CONFIG;
		}

		case "prod": {
			if (!ServerConfig.CG_PROD_CONFIG) {
				throw new Error(
					`No CG_PROD_CONFIG is available, yet a request to cg-prod was attempted?`
				);
			}

			return ServerConfig.CG_PROD_CONFIG;
		}
	}
}

function GetCGUrl(service: CGServices, cardInfo: CGCardInfo, game: CGSupportedGames) {
	const cgConf = GetCGConf(service);

	const params = new URLSearchParams();

	params.set("api_key", cgConf.API_KEY);
	params.set("card_id", cardInfo.cardID);
	params.set("pin", cardInfo.pin);

	return `${cgConf.URL}/${game}-scores?${params.toString()}`;
}
