import { ConvertFileS3, ParseDifficulty, ResolveS3Lamp } from "./converter";
import deepmerge from "deepmerge";
import CreateLogCtx from "lib/logger/logger";
import t from "tap";
import ResetDBState from "test-utils/resets";
import {
	LoadTachiIIDXData,
	MockParsedS3Score,
	Testing511Song,
	Testing511SPA,
} from "test-utils/test-data";
import type { S3Score } from "./types";

const logger = CreateLogCtx(__filename);

function cfile(data: S3Score) {
	return ConvertFileS3(data, {}, "file/solid-state-squad", logger);
}

t.test("#ConvertFileS3", (t) => {
	t.beforeEach(ResetDBState);
	t.beforeEach(() => {
		// @ts-expect-error hacky just-incase.
		delete MockParsedS3Score._id;
	});

	function mfile(merge: Partial<S3Score>) {
		return cfile(deepmerge(MockParsedS3Score, merge));
	}

	const dryScore = {
		game: "iidx",
		comment: null,
		importType: "file/solid-state-squad",
		service: "Solid State Squad",
		scoreData: {
			// percent: 6.36, -- fpa
			grade: "F",
			score: 100,
			lamp: "FULL COMBO",
			judgements: {
				pgreat: 25,
				great: 50,
				good: 0,
				bad: 0,
				poor: 4,
			},
			hitMeta: {},
		},
		scoreMeta: {},

		// can't be tested because the timestamp format doesnt specify a timezone.
		// timeAchieved: 1287460462000,
	};

	t.test("Should import a valid S3 score", async (t) => {
		const res = await cfile(MockParsedS3Score);

		t.hasStrict(
			res,
			{
				chart: Testing511SPA,
				song: Testing511Song,
				dryScore,
			},
			"Should correctly return the song, chart and DryScore."
		);

		t.end();
	});

	t.test("Should support comments in S3 scores", async (t) => {
		const res = await mfile({ comment: "FOO BAR" });

		t.hasStrict(
			res,
			{
				chart: Testing511SPA,
				song: Testing511Song,
				dryScore: deepmerge(dryScore, { comment: "FOO BAR" }),
			},
			"Should correctly return the song, chart and DryScore."
		);

		t.end();
	});

	t.test("Should find song case-insensitively", async (t) => {
		await LoadTachiIIDXData();

		const res = await mfile({ songname: "aBSolUte", diff: 7 });

		t.hasStrict(
			res,
			{
				chart: { songID: 97, difficulty: "HYPER", playtype: "SP" },
				song: { title: "ABSOLUTE" },

				// dryScore, dont care
			},
			"Should correctly return the song, chart and DryScore."
		);

		t.end();
	});

	t.test("Should reject invalid styles in S3 scores", (t) => {
		t.rejects(mfile({ styles: "3rd,4th,INVALID" }), {
			message: /Song has invalid style INVALID/u,
		});

		t.end();
	});

	t.test("Should throw ktdatanf if no song", (t) => {
		t.rejects(mfile({ songname: "INVALID SONG TITLE" }), {
			message: /Could not find song with title INVALID SONG TITLE/u,
		});

		t.end();
	});

	t.test("Should throw ktdatanf if no song", (t) => {
		t.rejects(mfile({ diff: "B" }), {
			message: /Could not find chart 5\.1\.1\. \(SP LEGGENDARIA/u,
		});

		t.end();
	});

	t.test("Should throw a skipscore if the song is 5key", (t) => {
		t.rejects(mfile({ diff: 5 }), {
			message: /5KEY scores are not supported/u,
		});

		t.end();
	});

	t.test("Should throw an invalidscore if the difficulty is invalid", (t) => {
		t.rejects(mfile({ diff: "INVALID" } as unknown as S3Score), {
			message: /Invalid difficulty INVALID/u,
		});

		t.end();
	});

	t.test("Should throw an invalidscore if the hardeasy is invalid", (t) => {
		t.rejects(
			mfile({ mods: { hardeasy: "INVALID" }, cleartype: "cleared" } as unknown as S3Score),
			{
				message: /Invalid hardeasy of INVALID while evaluating a 'cleared' score/u,
			}
		);

		t.end();
	});

	t.test("Should throw an invalidscore if the cleartype is invalid", (t) => {
		t.rejects(mfile({ cleartype: "INVALID" } as unknown as S3Score), {
			message: /Invalid cleartype of INVALID/u,
		});

		t.end();
	});

	t.test("Should throw an invalidscore if the exscore is greater than MAX", (t) => {
		t.rejects(mfile({ exscore: 10000 }), {
			message: /Invalid percent of 636/u,
		});

		t.end();
	});

	t.test("Should throw an invalidscore if the date is invalid.", (t) => {
		t.rejects(mfile({ date: "INVALID" }), {
			message: /Invalid\/Unparsable score timestamp of INVALID/u,
		});

		t.end();
	});

	t.end();
});

t.test("#ParseDifficulty", (t) => {
	t.beforeEach(ResetDBState);

	t.strictSame(ParseDifficulty("L7"), { playtype: "SP", difficulty: "NORMAL" });
	t.strictSame(ParseDifficulty(7), { playtype: "SP", difficulty: "HYPER" });
	t.strictSame(ParseDifficulty("A"), { playtype: "SP", difficulty: "ANOTHER" });
	t.strictSame(ParseDifficulty("B"), { playtype: "SP", difficulty: "LEGGENDARIA" });
	t.strictSame(ParseDifficulty("L14"), { playtype: "DP", difficulty: "NORMAL" });
	t.strictSame(ParseDifficulty(14), { playtype: "DP", difficulty: "HYPER" });
	t.strictSame(ParseDifficulty("A14"), { playtype: "DP", difficulty: "ANOTHER" });
	t.strictSame(ParseDifficulty("B14"), { playtype: "DP", difficulty: "LEGGENDARIA" });
	t.throws(() => ParseDifficulty(5));

	t.end();
});

t.test("#ResolveS3Lamp", (t) => {
	t.beforeEach(ResetDBState);

	t.equal(ResolveS3Lamp({ cleartype: "played" } as S3Score), "FAILED");
	t.equal(ResolveS3Lamp({ cleartype: "cleared", mods: {} } as S3Score), "CLEAR");
	t.equal(
		ResolveS3Lamp({ cleartype: "cleared", mods: { hardeasy: "E" } } as S3Score),
		"EASY CLEAR"
	);
	t.equal(
		ResolveS3Lamp({ cleartype: "cleared", mods: { hardeasy: "H" } } as S3Score),
		"HARD CLEAR"
	);
	t.equal(ResolveS3Lamp({ cleartype: "combo" } as S3Score), "FULL COMBO");
	t.equal(ResolveS3Lamp({ cleartype: "comboed" } as S3Score), "FULL COMBO");
	t.equal(ResolveS3Lamp({ cleartype: "perfect" } as S3Score), "FULL COMBO");
	t.equal(ResolveS3Lamp({ cleartype: "perfected" } as S3Score), "FULL COMBO");

	t.throws(() => ResolveS3Lamp({ cleartype: "invalid" } as unknown as S3Score));
	t.throws(() =>
		ResolveS3Lamp({ cleartype: "cleared", mods: { hardeasy: "invalid" } } as unknown as S3Score)
	);

	t.end();
});
