/* eslint-disable no-await-in-loop */
import db from "external/mongo/db";
import { rootLogger } from "lib/logger/logger";
import { CreateCalculatedData } from "lib/score-import/framework/calculated-data/calculated-data";
import { CreatePBDoc, UpdateChartRanking } from "lib/score-import/framework/pb/create-pb-doc";
import { CreateScoreID } from "lib/score-import/framework/score-importing/score-id";
import { CreateSessionCalcData } from "lib/score-import/framework/sessions/calculated-data";
import { ProcessScoreIntoSessionScoreInfo } from "lib/score-import/framework/sessions/sessions";
import { UpdateAllPBs } from "utils/calculations/recalc-scores";
import { FormatUserDoc, GetUserWithID } from "utils/user";
import type { KtLogger } from "lib/logger/logger";
import type { ScoreDocument } from "tachi-common";

/**
 * Updates a score from oldScore to newScore, applying all necessary state
 * changes on the way.
 *
 * @note You don't need to recalc the scoreID for newScore, it's done for you.
 */
export default async function UpdateScore(
	oldScore: ScoreDocument,
	newScore: ScoreDocument,
	updateOldChart = true
) {
	const userID = oldScore.userID;
	const user = await GetUserWithID(userID);

	if (!user) {
		rootLogger.severe(
			`User ${userID} does not exist, yet a score update was called for them? Panicking.`
		);
		throw new Error(
			`User ${userID} does not exist, yet a score update was called for them? Panicking.`
		);
	}

	const chartID = newScore.chartID;

	const chart = await db.charts[oldScore.game].findOne({
		chartID,
	});

	if (!chart) {
		rootLogger.severe(
			`Chart ${chartID} does not exist, yet a score update was called for it? Panicking.`
		);
		throw new Error(
			`Chart ${chartID} does not exist, yet a score update was called for it? Panicking.`
		);
	}

	// In the event that the new chart isn't under the same song as the previous one, the songID
	// needs to update.
	newScore.songID = chart.songID;

	const oldScoreID = oldScore.scoreID;

	const newScoreID = CreateScoreID(newScore.userID, newScore, newScore.chartID);

	// We need to change *so* many references to score IDs, and recalculate *so*
	// much stored state. Obviously, changing a scoreID is an exceptional circumstance
	// brought on by a bug.
	// So hopefully, we wont have to use this much.

	newScore.scoreID = newScoreID;

	const logger = rootLogger.child({
		context: ["Update Score", oldScore.scoreID, newScore.scoreID, FormatUserDoc(user)],
	}) as KtLogger;

	logger.verbose("Received Update Score request.");

	// eslint-disable-next-line require-atomic-updates
	newScore.calculatedData = await CreateCalculatedData(
		newScore,
		chart,
		newScore.scoreData.esd,
		logger
	);

	try {
		// Having _id defined will cause this to throw, causing it to not apply
		// the update.
		if (newScore._id) {
			logger.warn(
				`Passed a score with _id to UpdateScore. This property should not be set. Deleting this property and continuing anyway.`
			);

			// This property shouldn't be defined.
			delete newScore._id;
		}

		await db.scores.update(
			{
				scoreID: oldScoreID,
			},
			{ $set: newScore }
		);
	} catch (err) {
		logger.error(err);
		logger.warn(
			`Score ID ${newScoreID} already existed -- this update caused a collision. Removing old score and updating old references anyway.`
		);
		await db.scores.remove({
			scoreID: oldScoreID,
		});
	}

	const sessions = await db.sessions.find({
		"scoreInfo.scoreID": oldScoreID,
	});

	// another session already has the new score? (i.e. migrating to an already
	// existing score?)
	const existsElsewhere = await db.sessions.findOne({
		"scoreInfo.scoreID": newScoreID,
	});

	logger.verbose(`Updating ${sessions.length} sessions.`);

	// For every session that interacts with this score ID (there should only ever be one)
	for (const session of sessions) {
		const newScoreInfoArr = [];

		// Go over all the scoreInfo and alter the ones that involve this scoreID.
		for (const scoreInfo of session.scoreInfo) {
			let newScoreInfo = scoreInfo;

			// If this scoreInfo needs to be changed
			if (scoreInfo.scoreID === oldScoreID) {
				if (existsElsewhere) {
					// skip this, as this score already belongs to another session.
					continue;
				}

				if (newScore.timeAchieved === null) {
					// this shouldn't be in a session anymore.
					continue;
				}

				newScoreInfo.scoreID = newScoreID;

				// update scoreInfo
				const pbAsOfThen = await CreatePBDoc(
					newScore.userID,
					newScore.chartID,
					logger,
					newScore.timeAchieved
				);

				// @ts-expect-error we don't need the ranking data props. ignore this
				// err, it's nonsense.
				newScoreInfo = ProcessScoreIntoSessionScoreInfo(newScore, pbAsOfThen ?? null);
			}

			newScoreInfoArr.push(newScoreInfo);
		}

		const scores = await db.scores.find({
			scoreID: { $in: newScoreInfoArr.map((e) => e.scoreID) },
		});

		// update calculated data too.
		const newCalcData = CreateSessionCalcData(session.game, session.playtype, scores);

		await db.sessions.update(
			{
				sessionID: session.sessionID,
			},
			{
				$set: { scoreInfo: session.scoreInfo, calculatedData: newCalcData },
			}
		);
	}

	logger.verbose(`Updating PBs.`);

	// Update the PBs to reference properly.
	// We run updateAllPbs on just the modified chart -- the reason
	// for this is to update ranking info incase that might fall out of
	// sync as a result.
	await UpdateAllPBs([userID], {
		chartID: newScore.chartID,
	});

	await UpdateChartRanking(newScore.game, newScore.playtype, newScore.chartID);

	if (updateOldChart) {
		await UpdateAllPBs([userID], {
			chartID: oldScore.chartID,
		});
		await UpdateChartRanking(oldScore.game, oldScore.playtype, oldScore.chartID);
	}

	const imports = await db.imports.find({
		scoreIDs: oldScoreID,
	});

	logger.verbose(`Updating ${imports.length} imports.`);

	for (const importDoc of imports) {
		await db.imports.update(
			{
				importID: importDoc.importID,
			},
			{
				$set: {
					scoreIDs: importDoc.scoreIDs.map((e) => (e === oldScoreID ? newScoreID : e)),
				},
			}
		);
	}

	logger.verbose(`Done updating score.`);
}
